/***********************************************************************

input.c - read and process input lines

This file is a part of sscsss -- Simple Single Cycle Sample Set Synthesizer.

Copyright (C) 2002 David Baker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "sscsss.h"
#include "s1m.h"
#include "proto.h"

/**********************************************************************

get_input

***********************************************************************/

#define LINESIZE 120 

static char line[LINESIZE];
static char arg1[LINESIZE], arg2[LINESIZE], arg3[LINESIZE], arg4[LINESIZE],
	    arg5[LINESIZE], arg6[LINESIZE], arg7[LINESIZE], arg8[LINESIZE];
static char *argsv[8];
static int  argc;

int get_input(M_P m)
{
    for ( ; ; ) {
	if ( fgets(line, LINESIZE, m->infile) == NULL ) return 0;
	(m->linecnt)++;
	/* should have more '%'s than needed to check for too many */
	argc = sscanf(line,"%s%s%s%s%s%s%s%s",
	  arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
	if ( argc < 1 ) continue;		/* blank line */
	if ( argc == -1 ) continue;		/* blank line */
	if ( line[0] == '#' ) continue;		/* comment line */
	if ( arg1[strlen(arg1)-1] == ':' )	/* remove : from end of arg1 */
	  arg1[strlen(arg1)-1] = '\0';
	break;
    }
    return 1;
}

/**********************************************************************

get_opts_fname, also initializes temporaries in m

***********************************************************************/

/* forward references */
void option_proc(M_P);
void file_proc(M_P, char *);

void get_opts_fname(M_P m, char filename[])
{
    /* initialize argsv[] */
    argsv[0] = arg1; argsv[1] = arg2; argsv[2] = arg3; argsv[3] = arg4;
    argsv[4] = arg5; argsv[5] = arg6; argsv[6] = arg7; argsv[7] = arg8;

    /* initialize counters etc. */
    m->v_waves   = 0;		/* number of waves in current voice  */
    m->w_samples = 0;		/* number of samples in current wave */
    m->h_no      = 0;		/* harmonic number in current sample */
    m->s_type	 = SC_NONE;	/* parameter being scaled	     */

    /* initialize list backs */
    m->ws_back  = NULL;	/* pointer to back of current set list	     */
    m->o_back   = NULL;	/* pointer to back of current offset list    */
    m->fi_back  = NULL;	/* pointer to back of current filter list    */
    m->sc_back  = NULL;	/* pointer to back of current scale list     */
    m->gp_back  = NULL; /* pointer to back of current graph point list */
    m->fo_back  = NULL;	/* pointer to back of current formant list   */
    m->wp_back  = NULL;	/* pointer to back of current split list     */
    m->n_back   = NULL;	/* pointer to back of current note list      */
    m->wpc_back = NULL;	/* pointer to back of current condition list */
    m->w_back   = NULL;	/* pointer to back of current wave list	     */
    m->s_back   = NULL;	/* pointer to back of current sample list    */
    m->h_back   = NULL;	/* pointer to back of current harmonic list  */

    for ( ; ; ) {

	/* state IS_INIT, initial */
	if ( m->state == IS_INIT ) {
	    /* nothing to process yet */

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for file.");
	    
	    if ( ! strcmp(argsv[0], "file") ) m->state = IS_FILE;
	    else errex(m, 1, "Expecting file.");
	}

	/* state IS_FILE, file */
	else if ( m->state == IS_FILE ) {
	    file_proc(m, filename);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for option or voice.");
	    
	    if      ( ! strcmp(argsv[0], "option") ) m->state = IS_OPTION;
	    else if ( ! strcmp(argsv[0], "voice" ) ) {
		m->state = IS_VOICE;
		return;
	    }
	    else errex(m, 1, "Expecting option or voice.");
	}

	/* state IS_OPTION, option */
	else if ( m->state == IS_OPTION ) {
	    option_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for option or voice.");

	    if      ( ! strcmp(argsv[0], "option") ) m->state = IS_OPTION;
	    else if ( ! strcmp(argsv[0], "voice" ) ) {
		m->state = IS_VOICE;
		return;
	    }
	    else errex(m, 1, "Expecting option or voice.");
	}
    }
}

/***********************************************************************

file_proc

***********************************************************************/

void file_proc(M_P m, char *filename)
{
    char logfilename[13];

    /* process file */
    if ( argc != 2 ) errex(m, 1, "Must have (only) a filename.");
    if ( strlen(arg2) > 8 ) errex(m, 1, "Filename length must be <= 8.");
    if ( ! strcmp(arg2, "init")) errex(m, 1, "Filename cannot be init.");
    strcpy(filename, arg2);

    /* construct log file name and open */
    strcpy(logfilename, filename);
    strcat(logfilename, ".log");
    m->log = fopen(logfilename, "w");
    if ( m->log == NULL ) errex(m, 1, "Cannot open log file.");
}

/***********************************************************************

option_proc

***********************************************************************/

void option_proc(M_P m)
{
    int    ret;
    double d_arg;

    /* check number of args */
    if ( argc < 2 )
      errex(m, 1, "Option must have at least one more argument.");

    /* change lowest frequency accepted */
    if      ( ! strcmp(arg2, "lowfreq") ) {
	if ( argc != 3 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(arg3, &d_arg);
	if ( ret ) errex(m, 1, "Lowfreq bad number.");
	if ( d_arg < .01 ) errex(m, 1, "Lowfreq must be >= 0.01");
	if ( d_arg > 8 )   errex(m, 1, "Lowfreq must be <= 8");
	/* ok, set low frequency to new value */
	m->o_lowfreq = d_arg;
    }

    /* change lowest dba generated */
    else if ( ! strcmp(arg2, "mindba") ) {
	if ( argc != 3 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(arg3, &d_arg);
	if ( ret ) errex(m, 1, "Mindba bad number.");
	if ( d_arg > -40 ) errex(m, 1, "Mindba must be <= -40");
	/* ok, set min dba to new value */
	m->o_mindba = d_arg;
    }

    /* debug */
    else if ( ! strcmp(arg2, "debug") ) {
	if ( argc > 2 )
	  errex(m, 1, "Option debug has no additional arguments.");
	/* ok, set debug on */
	m->o_debug = 1;
    }

    /* log generated wave set */
    else if ( ! strcmp(arg2, "logws") ) {
	if ( argc > 2 )
	  errex(m, 1, "Option logws has no additional arguments.");
	/* ok, set log wave set on */
	m->o_logws = 1;
    }

    /* make plot files */
    else if ( ! strcmp(arg2, "mkplot") ) {
	if ( argc > 2 )
	  errex(m, 1, "Option mkplot has no additional arguments.");
	/* ok, set mkplot on */
	m->o_mkplot = 1;
    }

    /* plot frequency */
    else if ( ! strcmp(arg2, "plotfreq") ) {
	if ( argc > 2 )
	  errex(m, 1, "Option plotfreq has no additional arguments.");
	/* ok, set plotfreq on */
	m->o_plotfreq = 1;
    }

    /* add pause command to end of plot file */
    else if ( ! strcmp(arg2, "plotpause") ) {
	if ( argc > 2 )
	  errex(m, 1, "Option plotpause has no additional arguments.");
	/* ok, set plotpause on */
	m->o_plotpause = 1;
    }

    /* error */
    else errex(m, 1, "Unknown option.");
}

/***********************************************************************

get_voice

***********************************************************************/

/* forward references from get_voice, for top down reading */

void voice_proc(M_P);
void set_proc(M_P);
void offset_proc(M_P);
void filter_proc(M_P);
void scale_proc(M_P);
void point_proc(M_P);
void gpoint_proc(M_P);
void formant_proc(M_P);
void split_proc(M_P);
void note_proc(M_P);
void condition_proc(M_P);
void wave_proc(M_P);
void sample_proc(M_P);
void harmonic_proc(M_P);

void get_voice(M_P m)
{
    for ( ; ; ) {

	/* state IS_VOICE, voice */
	if ( m->state == IS_VOICE ) {
	    voice_proc(m);

	    if ( ! get_input(m) )  /* if end of file */
	      errex(m, 1, "End of file while looking for set or wave.");

	    if      (!strcmp(arg1, "set" )) m->state=IS_SET;
	    else if (!strcmp(arg1, "wave")) m->state=IS_WAVE;
	    else errex(m, 1, "Expecting set or wave.");
	}

	/* state IS_SET, set */
	else if ( m->state == IS_SET ) {
	    set_proc(m);
	    
	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for offset, filter, "
	        "formant, or split.");

	    if      (!strcmp(arg1, "offset" )) m->state=IS_OFFSET;
	    else if (!strcmp(arg1, "filter" )) m->state=IS_FILTER;
	    else if (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
	    else if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
	    else errex(m, 1, "Expecting offset, filter, formant or split.");
	}

	/* state IS_OFFSET, offset */
	else if ( m->state == IS_OFFSET ) {
	    offset_proc(m);

	    if ( ! get_input(m) )
	      errex(m,1,"End of file while looking for offset, scale, filter, "
	        "formant, or split.");

	    if      (!strcmp(arg1, "offset")) m->state=IS_OFFSET;
	    else if (!strcmp(arg1, "scale" )) m->state=IS_OFFSET_SC;
	    else {
		m->o_back = NULL;
		if      (!strcmp(arg1, "filter" )) m->state=IS_FILTER;
		else if (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
		else if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
		else errex(m, 1,
		  "Expecting offset, scale, filter, formant or split.");
	    }
	}

	/* state IS_OFFSET_SC, scale (offset) */
	else if ( m->state == IS_OFFSET_SC ) {
	    scale_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for point.");

	    if (!strcmp(arg1, "point")) m->state=IS_OFFSET_SC_P;
	    else errex(m, 1, "Expecting point.");
	}

	/* state IS_OFFSET_SC_P, point (offset) */
	else if ( m->state == IS_OFFSET_SC_P ) {
	    point_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for point, scale, offset,"
		"filter, formant, or split.");

	    if      (!strcmp(arg1, "point")) m->state=IS_OFFSET_SC_P;
	    else {
		m->sc_back = NULL;
		m->s_type  = SC_NONE;
		if      (!strcmp(arg1, "scale" )) m->state=IS_OFFSET_SC;
		else if (!strcmp(arg1, "offset")) m->state=IS_OFFSET;
		else {
		    m->o_back = NULL;
		    if      (!strcmp(arg1,"filter" )) m->state=IS_FILTER;
		    else if (!strcmp(arg1,"formant")) m->state=IS_FORMANT;
		    else if (!strcmp(arg1,"split"  )) m->state=IS_SPLIT;
		    else errex(m, 1, "Expecting point, scale, offset, filter, "
		      "formant, or split.");
		}
	    }
	}

	/* state IS_FILTER, filter */
	else if ( m->state == IS_FILTER ) {
	    filter_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for filter, scale, "
	      "gpoint, formant, or split.");

	    if      (!strcmp(arg1, "filter")) m->state=IS_FILTER;
	    else if (!strcmp(arg1, "scale" )) m->state=IS_FILTER_SC;
	    else if (!strcmp(arg1, "gpoint")) m->state=IS_FILTER_GP;
	    else {
		m->fi_back = NULL;
		if      (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
		else if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
		else errex(m, 1,
		  "Expecting filter, scale, gpoint, formant or split.");
	    }
	}

	/* state IS_FILTER_SC, scale (filter) */
	else if ( m->state == IS_FILTER_SC ) {
	    scale_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for point.");

	    if      (!strcmp(arg1, "point")) m->state=IS_FILTER_SC_P;
	    else errex(m, 1, "Expecting point.");
	}

	/* state IS_FILTER_SC_P, point */
	else if ( m->state == IS_FILTER_SC_P ) {
	    point_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for point, filter, "
	      "formant, or split.");

	    if      (!strcmp(arg1, "point")) m->state=IS_FILTER_SC_P;
	    else {
		m->sc_back = NULL;
		m->s_type  = SC_NONE;
		if      (!strcmp(arg1, "scale" )) m->state=IS_FILTER_SC;
		else if (!strcmp(arg1, "filter")) m->state=IS_FILTER;
		else {
		    m->fi_back = NULL;
		    if      (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
		    else if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
		    else errex(m, 1, "Expecting point, scale, filter, "
		      "formant, or split.");
		}
	    }
	}

	/* state IS_FILTER_GP, graph point */
	else if ( m->state == IS_FILTER_GP ) {
	    gpoint_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for gpoint, filter,"
		"formant, or split.");

	    if ( ! strcmp(arg1, "gpoint")) m->state = IS_FILTER_GP;
	    else {
		m->gp_back = NULL;
		if (!strcmp(arg1, "filter")) m->state=IS_FILTER;
		else {
		    m->fi_back = NULL;
		    if      (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
		    else if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
		    else errex(m, 1, "Expecting gpoint, filter, "
		      "formant, or split.");
		}
	    }
	}

	/* state IS_FORMANT, formant */
	else if ( m->state == IS_FORMANT ) {
	    formant_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1,
	       "End of file while looking for formant, gpoint, or split.");

	    if (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
	    else if (!strcmp(arg1, "gpoint")) m->state=IS_FORMANT_GP;
	    else {
		m->fo_back = NULL;
		if (!strcmp(arg1, "split")) m->state=IS_SPLIT;
		else errex(m, 1,
		  "Expecting formant, gpoint, or split.");
	    }
	}

	/* state IS_FORMANT_GP, graph point */
	else if ( m->state == IS_FORMANT_GP ) {
	    gpoint_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for gpoint, formant,"
	       " or split.");

	    if ( ! strcmp(arg1, "gpoint")) m->state = IS_FORMANT_GP;
	    else {
		m->gp_back = NULL;
		if (!strcmp(arg1, "formant")) m->state=IS_FORMANT;
		else {
		    m->fo_back = NULL;
		    if (!strcmp(arg1, "split"  )) m->state=IS_SPLIT;
		    else errex(m, 1, "Expecting gpoint, formant, or split.");
		}
	    }
	}

	/* state IS_SPLIT, split */
	else if ( m->state == IS_SPLIT ) {
	    split_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for note or condition.");

	    if      (!strcmp(arg1, "note"     )) m->state=IS_NOTE;
	    else if (!strcmp(arg1, "condition")) m->state=IS_SPLIT_C;
	    else errex(m, 1, "Expecting note or condition.");
	}

	/* state IS_NOTE, note */
	else if ( m->state == IS_NOTE ) {
	    note_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for note or condition.");

	    if      (!strcmp(arg1, "note"     )) m->state=IS_NOTE;
	    else if (!strcmp(arg1, "condition")) {
		m->n_back = NULL;
		m->state  = IS_SPLIT_C;
	    }
	    else errex(m, 1, "Expecting note or condition.");
	}

	/* state IS_SPLIT_C, condition */
	else if ( m->state == IS_SPLIT_C ) {
	    condition_proc(m);

	    if ( ! get_input(m) ) {
		gen_waves(m);
		m->state = IS_EOF;
		return;
	    }

	    if (!strcmp(arg1, "condition")) m->state=IS_SPLIT_C;
	    else {
		m->wpc_back = NULL;
		if (!strcmp(arg1, "split")) m->state=IS_SPLIT;
		else {
		    m->wp_back = NULL;
		    if (!strcmp(arg1, "set")) m->state=IS_SET;
		    else {
			m->ws_back = NULL;
			if      (!strcmp(arg1, "wave" )) {
			    gen_waves(m);
			    m->state=IS_WAVE;
			}
			else if (!strcmp(arg1, "voice")) {
			    gen_waves(m);
			    m->w_back = NULL;
			    m->state = IS_VOICE;
			    return;
			}
			else errex(m, 1, "Expecting condition, split, wave, "
			  "set, voice, or end of file.");
		    }
		}
	    }
	}

	/* state IS_WAVE, wave */
	else if ( m->state == IS_WAVE ) {
	    wave_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for sample.");

	    if (!strcmp(arg1, "sample")) m->state=IS_SAMPLE;
	    else errex(m, 1, "Expecting sample.");
	}

	/* state IS_SAMPLE, sample */
	else if ( m->state == IS_SAMPLE ) {
	    sample_proc(m);

	    if ( ! get_input(m) )
	      errex(m, 1, "End of file while looking for harmonic.");

	    if (!strcmp(arg1, "harmonic")) m->state=IS_HARMONIC;
	    else errex(m, 1, "Expecting harmonic.");
	}

	/* state IS_HARMONIC, harmonic */
	else if ( m->state == IS_HARMONIC ) {
	    harmonic_proc(m);

	    if ( ! get_input(m) ) {
		m->w_back->layers = m->w_samples;
		m->state = IS_EOF;
		return;
	    }

	    if (!strcmp(arg1, "harmonic")) m->state=IS_HARMONIC;
	    else {
		m->h_back = NULL;
		if (!strcmp(arg1, "sample")) m->state=IS_SAMPLE;
		else {
		    m->s_back = NULL;
		    m->w_back->layers = m->w_samples;
		    m->w_samples = 0;
		    if (!strcmp(arg1, "wave")) m->state=IS_WAVE;
		    else {
			m->w_back = NULL;
			if (!strcmp(arg1, "voice")) {
			    m->state = IS_VOICE;
			    return;
			}
			else errex(m, 1, "Expecting harmonic, sample, wave, "
			  "voice, or end of file.");
		    }
		}
	    }
	}
    }
}

/***********************************************************************

voice_proc

***********************************************************************/

void voice_proc(M_P m)
{
    int ret, bend, low, high, norm;

    /* check number of args */
    if ( argc < 3 || argc > 7 ) errex(m, 1, "Incorrect number of arguments.");

    /* check name length */
    if ( strlen(argsv[1]) > 12 ) errex(m, 1, "Name > 12 chars.");

    /* check category */
    if ( chkcat(argsv[2]) == -1 ) errex(m, 1, "Unknown Category.");

    /* check optional bend, required for wave sets */
    if ( argc > 3 ) {
	ret = aconvi(argsv[3], &bend);
	if ( ret ) errex(m, 1, "Bend bad number.");
	if ( bend<0 || bend>127 ) errex(m, 1, "Bend must be >= 0 and < 128.");
    }
    else bend = 0;

    /* check optional low and high, required for wave sets */
    if ( argc > 4 ) {
	if ( argc < 6 ) errex(m, 1, "Must specify both low and high.");
	ret = aconvi(argsv[4], &low);
	if ( ret ) errex(m, 1, "Low bad number.");
	if ( low < 0 || low > 127 ) errex(m, 1, "Low must be >= 0 and < 128.");
	ret = aconvi(argsv[5], &high);
	if ( ret ) errex(m, 1, "High bad number.");
	if ( high<0 || high>127 ) errex(m, 1, "High must be >= 0 and < 128.");
	if ( low > high ) errex(m, 1, "Low must be <= High.");
    }
    else {
	low  = -1;
	high = -1;
    }

    /* check optional normalization type */
    if ( argc > 6 ) {
	if      ( ! strcmp(argsv[6], "sample") ) norm = 1;
	else if ( ! strcmp(argsv[6], "voice" ) ) norm = 2;
	else errex(m, 1, "Normalization type must be sample or voice.");
    }
    else norm = 0;

    /* input ok, update m, no voice list */
    strcpy(m->v_cat,  argsv[2]);
    strcpy(m->v_name, argsv[1]);
    m->v_bend    = bend;
    m->v_low     = low;
    m->v_high    = high;
    m->v_norm	 = norm;
    m->v_waves   = 0;
    m->w_samples = 0;
}

/***********************************************************************

set_proc

***********************************************************************/

void set_proc(M_P m)
{
    WS_P   ws_ptr;

    /* check for voice low high */
    if ( m->v_low == -1 )
	errex(m, 1, "Voice high/low notes needed for defining wave sets.");

    /* check number of args */
    if ( argc != 1 ) errex(m, 1, "Incorrect number of arguments.");

    /* input ok */

    /* allocate new wave set structure and put in list */
    ws_ptr = (WS_P) malloc(sizeof(WS_S));
    (m->d_malloc)++;
    if ( ws_ptr == NULL ) errex(m, 1,"Unable to allocate wave set struct.");
    if ( m->ws_back == NULL ) m->ws_list = ws_ptr;
    else m->ws_back->next = ws_ptr;
    m->ws_back = ws_ptr;

    /* set data in structure */
    ws_ptr->o_list	= NULL;
    ws_ptr->fi_list	= NULL;
    ws_ptr->fo_list	= NULL;
    ws_ptr->wp_list	= NULL;
    ws_ptr->next	= NULL;

    if ( m->v_norm == 0 ) m->v_norm = 2; /* if set used, default to voice */
}

/***********************************************************************

offset_proc

***********************************************************************/

void offset_proc(M_P m)
{
    int    ret, h_no;
    double level, phase;
    O_P ptr;

    /* check number of args */
    if ( argc < 3 || argc > 4 ) errex(m, 1, "Incorrect number of arguments.");

    /* check harmonic number */
    ret = aconvi(argsv[1], &h_no);
    if ( ret ) errex(m, 1, "Harmonic bad number.");
    if ( h_no < 1 ) errex(m, 1, "Harmonic must be >= 1.");

    /* check level */
    ret = aconvd(argsv[2], &level);
    if ( ret ) errex(m, 1, "Level bad number.");
    if ( level < -90.0 || level > 90.0 )
      errex(m, 1, "Level must be between -90 and 90.");

    /* check optional phase */
    if ( argc == 4 ) {	/* phase specified */
	ret = aconvd(argsv[3], &phase);
	if ( ret ) errex(m, 1, "Phase bad number.");
	if ( phase < 0.0 || phase >= 1.0 )
	    errex(m, 1, "Phase must be >= 0.0 and < 1.0 .");
    }
    else phase = 0.0; /* phase not specified, set to 0.0 */

    /* input ok */

    /* allocate offset struct and put in list */
    ptr = (O_P) malloc (sizeof(O_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 1, "Unable to allocate offset struct.");
    if ( m->o_back == NULL ) m->ws_back->o_list = ptr;
    else (m->o_back)->next = ptr;
    m->o_back = ptr;
    
    /* set data in structure */
    ptr->no		= h_no;
    ptr->level		= level;
    ptr->phase		= phase;
    ptr->level_s_list	= NULL;
    ptr->phase_s_list	= NULL;
    ptr->next		= NULL;
}

/***********************************************************************

filter_proc

***********************************************************************/

void filter_proc(M_P m)
{
    double p1, p2, p3, p4;
    int    ret, type;
    FI_P f_ptr;

    /* check number of args */
    if ( argc < 2 ) errex(m, 1, "Incorrect number of arguments.");

    /* check type */
    if      ( ! strcmp(argsv[1], "cos"      ) ) type = F_COS;
    else if ( ! strcmp(argsv[1], "triangle" ) ) type = F_TRI;
    else if ( ! strcmp(argsv[1], "highpass" ) ) type = F_HPF;
    else if ( ! strcmp(argsv[1], "lowpass"  ) ) type = F_LPF;
    else if ( ! strcmp(argsv[1], "lowpass_f") ) type = F_LPF_F;
    else if ( ! strcmp(argsv[1], "highshelf") ) type = F_HS;
    else if ( ! strcmp(argsv[1], "lowshelf" ) ) type = F_LS;
    else if ( ! strcmp(argsv[1], "band"     ) ) type = F_BAND;
    else if ( ! strcmp(argsv[1], "graph"    ) ) type = F_GRAPH;
    else errex(m, 1, "Unknown filter type.");

    /* check p1 */
    if ( type != F_GRAPH ) {
	ret = aconvd(argsv[2], &p1);
	if ( ret ) errex(m, 1, "P1, bad number.");
	if ( p1 < 0.0 || p1 > 10.0 )
	  errex(m, 1, "P1 not between 0.0 and 10.0 .");
    }
    else {
	if ( argc > 2 ) errex(m, 1, "Too many arguments for this type.");
	p1 = 0.0; p2 = 0.0; p3 = 0.0; p4 = 0.0;
    }

    /* check p2 */
    if ( type != F_GRAPH ) {
	ret = aconvd(argsv[3], &p2);
	if ( ret ) errex(m, 1, "P2 bad number.");
	switch (type) {
	    case F_COS: case F_TRI:  case F_HS: case F_LS:
		if ( p2 < -90.0 || p2 > 90.0 )
		  errex(m, 1, "P2 not between -90.0 and 90.0 .");
		break;
	    case F_HPF: case F_LPF: case F_LPF_F:
		if ( p2 == 0.0 )
		  errex(m, 1, "P2 may not be 0.0 .");
		break;
	    case F_BAND:
		if ( p2 < 0.0 || p2 > 10.0 )
		  errex(m, 1, "P2 not between 0.0 and 10.0 .");
		break;
	}
    }
    else {
	if ( argc > 3 ) errex(m, 1, "Too many arguments for this type.");
	p1 = 0.0; p2 = 0.0; p3 = 0.0; p4 = 0.0;
    }

    /* check p3, if needed */
    if ( type == F_COS || type == F_TRI || type == F_HS || type == F_LS ||
      type == F_BAND ) {
	if ( argc < 5 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(argsv[4], &p3);
	if ( ret ) errex(m, 1, "P3 bad number.");
	switch (type) {
	    case F_COS: case F_TRI:
		if ( p3 < 0.0 || p3 > 10.0 )
		  errex(m, 1, "P3 not between 0.0 and 10.0 .");
		break;
	    case F_HS:  case F_LS:
		if ( p3 == 0.0 )
		  errex(m, 1, "P3 may not be 0.0 .");
		if ( ( p2 < 0 && p3 > 0 ) || ( p2 > 0 && p3 < 0 ) )
		  errex(m, 1, "p2 and p3 must have same sign.");
		break;
	    case F_BAND:
		if ( p3 < -90.0 || p3 > 90.0 )
		  errex(m, 1, "P3 not between -90.0 and 90.0 .");
		break;
	}
    }
    else if ( argc > 4 ) errex(m, 1, "Too many arguments for this type.");
    else p3 = 0.0; /* not used */

    /* check p4, if needed */
    if ( type == F_BAND ) {
	if ( argc < 6 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(argsv[5], &p4);
	if ( ret ) errex(m, 1, "P4 bad number.");
	switch (type) {
	    case F_BAND:
		if ( p4 == 0.0 )
		  errex(m, 1, "P4 may not be 0.0 .");
		break;
	}
    }
    else if ( argc > 5 ) errex(m, 1, "Too many arguments for this type.");
    else p4 = 0.0; /* not used */

    /* input ok */

    /* allocate new filter struct and put in list */
    f_ptr = (FI_P) malloc(sizeof(FI_S));
    (m->d_malloc)++;
    if ( f_ptr == NULL ) errex(m, 1,"Unable to allocate filter struct.");
    if ( m->fi_back == NULL ) m->ws_back->fi_list = f_ptr;
    else m->fi_back->next = f_ptr;
    m->fi_back = f_ptr;

    /* set data in structure */
    f_ptr->type      = type;
    f_ptr->p1        = p1;
    f_ptr->p2        = p2;
    f_ptr->p3        = p3;
    f_ptr->p4        = p4;
    f_ptr->p1_s_list = NULL;
    f_ptr->p2_s_list = NULL;
    f_ptr->p3_s_list = NULL;
    f_ptr->p4_s_list = NULL;
    f_ptr->gp_list   = NULL;
    f_ptr->next      = NULL;
}

/***********************************************************************

scale_proc

***********************************************************************/

void scale_proc(M_P m)
{
    /* check for scaling filter type graph, not permitted */
    if ( m->state == IS_FILTER_SC && m->fi_back->type == F_GRAPH )
      errex(m, 1, "Cannot scale filter type graph.");

    /* check number of arguments */
    if ( argc != 2 ) errex(m, 1, "Incorrect number of arguments.");
	    if ( m->s_type == SC_NONE ) {/* if starting new point list */
    }

    /* check attribute, depends on state */
    if      ( m->state ==  IS_OFFSET_SC ) {	/* scaling offset */
	if      ( ! strcmp(argsv[1], "level"  ) ) {
	    if ( m->o_back->level_s_list != NULL ) 
	      errex(m, 1, "Level for this offset already scaled.");
	    m->s_type = SC_O_LEVEL;	/* now scaling offset level */
	}
	else if ( ! strcmp(argsv[1], "phase") ) {
	    if ( m->o_back->phase_s_list != NULL ) 
	      errex(m, 1, "Phase for this offset already scaled.");
	    m->s_type = SC_O_PHASE;	/* now scaling offset phase */
	}
	else errex(m, 1, "Can only scale level or phase for an offset.");
    }
    else if ( m->state == IS_FILTER_SC ) {	/* scaling filter */
	if      ( ! strcmp(argsv[1], "p1") ) {
	    if ( m->fi_back->p1_s_list != NULL ) 
	      errex(m, 1, "P1 for this filter already scaled.");
	    m->s_type = SC_F_P1; /* now scaling filter p1 */
	}
	else if ( ! strcmp(argsv[1], "p2"  ) ) {
	    if ( m->fi_back->p2_s_list != NULL ) 
	      errex(m, 1, "P2 for this filter already scaled.");
	    m->s_type = SC_F_P2;	/* now scaling filter p2 */
	}
	else if ( ! strcmp(argsv[1], "p3" ) ) {
	    if ( m->fi_back->type == F_LPF ||
		 m->fi_back->type == F_LPF_F ||
		 m->fi_back->type == F_HPF )
		errex(m, 1, "No p3 for this filter type.");
	    if ( m->fi_back->p3_s_list != NULL ) 
	      errex(m, 1, "P3 for this filter already scaled.");
	    m->s_type = SC_F_P3;	/* now scaling filter p3 */
	}
	else if ( ! strcmp(argsv[1], "p4" ) ) {
	    if ( m->fi_back->type != F_BAND )
		errex(m, 1, "No p4 for this filter type");
	    if ( m->fi_back->p4_s_list != NULL ) 
	      errex(m, 1, "P4 for this filter already scaled.");
	    m->s_type = SC_F_P4;	/* now scaling filter p4 */
	}
	else errex(m, 1,
	  "Can only scale p1, p2, p3, p4 for a filter.");
    }
    else errex(m, 1, "Bug! scale_proc called with unknown state.");
}

/***********************************************************************

point_proc

***********************************************************************/

void point_proc(M_P m)
{
    double value, freq, prev_freq;
    int ret;
    SC_P ptr;

    /* check number of args */
    if ( argc != 3 ) errex(m, 1, "Must have point value frequency.");

    /* check frequency > previous point or 1st harmonic of lowest sample */
    ret = aconvd(argsv[2], &freq);
    if ( ret ) errex(m, 1, "Frequency bad number.");
    
    if ( m->sc_back != NULL ) 	/* if there is a previous point	    */
	prev_freq = m->sc_back->freq;
    else /* if no previous point, use frequency of lowest note in voice */
	prev_freq = 440.0 * pow(pow(2.0, 1.0 / 12.0), m->v_low - 69);

    if ( freq <= prev_freq ) errex(m, 1, "Frequency must be increasing.");

    /* check value (depends on parameter being scaled */
    ret = aconvd(argsv[1], &value);
    if ( ret ) errex(m, 1, "Value bad number.");

    if      ( m->s_type ==  SC_O_LEVEL ) {	/* if offset level	*/
	if ( value < -90.0 || value > 90.0 ) errex(m, 1,
	  "Level must be between -90 and 90.");
    }
    else if ( m->s_type ==  SC_O_PHASE ) {	/* if offset phase	*/
	if ( value < 0.0 || value >= 1.0 ) errex(m, 1,
	  "Phase must be >= than 0.0 and < 1.0 .");
    }
    else if ( m->s_type == SC_F_P1 ) {	/* if filter p1	*/
	if ( value < 0.0 || value > 10.0 ) errex(m, 1,
	  "P1 must be between 0 and 10.");
    }
    else if ( m->s_type == SC_F_P2 ) {	/* if filter p2	*/
	switch ( m->fi_back->type ) {
	    case F_COS: case F_TRI: case F_HS: case F_LS:
		if ( value < -90.0 || value > 90.0 )
		  errex(m, 1, "P2 must be between -90 and 90.");
		break;
	    case F_HPF: case F_LPF: case F_LPF_F:
		if ( value == 0.0 )
		  errex(m, 1, "P2 may not be 0.0 .");
		break;
	    case F_BAND:
		if ( value < 0.0 || value > 10.0 )
		  errex(m, 1, "P2 not between 0.0 and 10.0 .");
		break;
	}
    }
    else if ( m->s_type == SC_F_P3 ) {	/* if filter p3	*/
	switch ( m->fi_back->type ) {
	    case F_COS: case F_TRI:
		if ( value <= 0.0 || value > 10.0 )
		  errex(m, 1, "P3 must be > 0 and <= 10.");
		break;
	    case F_HS: case F_LS:
		if ( value == 0.0 )
		  errex(m, 1, "P3 may not be 0.0 .");
		break;
	    case F_BAND:
		if ( value < -90.0 || value > 90.0 )
		  errex(m, 1, "P3 must be between -90.0 and 90.0 .");
		break;
	}
    }
    else if ( m->s_type == SC_F_P4 ) {	/* if filter P4	*/
	if ( value == 0.0 )
	  errex(m, 1, "P4 may not be 0.0 .");
    }

    /* input ok */

    /* allocate new point structure and put in correct list */
    ptr = (SC_P) malloc(sizeof(SC_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 1, "Unable to allocate scale struct.");
    if ( m->sc_back == NULL ) {
	if      ( m->s_type == SC_O_LEVEL) m->o_back->level_s_list = ptr;
	else if ( m->s_type == SC_O_PHASE) m->o_back->phase_s_list = ptr;
	else if ( m->s_type == SC_F_P1 )   m->fi_back->p1_s_list   = ptr;
	else if ( m->s_type == SC_F_P2 )   m->fi_back->p2_s_list   = ptr;
	else if ( m->s_type == SC_F_P3 )   m->fi_back->p3_s_list   = ptr;
	else if ( m->s_type == SC_F_P4 )   m->fi_back->p4_s_list   = ptr;
	else errex(m, 1,
	  "Bug! m->s_type unknown when allocating point.");

    }
    else m->sc_back->next = ptr;
    m->sc_back = ptr;

    /* set data in struct */
    ptr->value = value;
    ptr->freq  = freq;
    ptr->next  = NULL;
}

/***********************************************************************

gpoint_proc

***********************************************************************/

void gpoint_proc(M_P m)
{
    double freq, gain;
    int ret;
    GP_P g_ptr;

    /* check for graph type filter/formant */
    if ( ( m->state == IS_FILTER_GP  && m->fi_back->type != F_GRAPH ) ||
	 ( m->state == IS_FORMANT_GP && m->fo_back->type != F_GRAPH ) )
      errex(m, 1, "gpoint only allowed for graph filter/formant.");

    /* check number of args */
    if ( argc != 3 ) errex(m, 1, "Incorrect number of arguments.");

    /* convert args, and check limits */
    ret = aconvd(argsv[1], &freq);
    if ( ret ) errex(m, 1, "Frequency bad number.");
    if ( m->state == IS_FILTER_GP && ( freq < 0.0 || freq > 11.0 ) )
      errex(m, 1, "Frequency (octave) not between 0 and 11.");
    else if ( freq < 0 || freq > 40000 )
      errex(m, 1, "Frequency not between 0 and 40000.");
    
    ret = aconvd(argsv[2], &gain);
    if ( ret ) errex(m, 1, "Gain bad number.");
    if ( gain < -90.0 || gain > 90.0 )
      errex(m, 1, "Gain not between -90 and 90.");

    /* check for increasing frequency */
    if ( m->gp_back != NULL && m->gp_back->freq >= freq )
      errex(m, 1, "Frequencies must be increasing.");

    /* allocate new graph point structure */
    g_ptr = (GP_P) malloc(sizeof(GP_S));
    (m->d_malloc)++;
    if ( g_ptr == NULL ) errex(m, 1,"Unable to allocate graph point struct.");
    if ( m->gp_back == NULL ) {
	if ( m->state == IS_FILTER_GP ) m->fi_back->gp_list = g_ptr;
	else                            m->fo_back->gp_list = g_ptr;
    }
    else m->gp_back->next = g_ptr;
    m->gp_back = g_ptr;

    /* set data in struct */
    g_ptr->freq = freq;
    g_ptr->gain = gain;
    g_ptr->next = NULL;
}

/***********************************************************************

formant_proc

***********************************************************************/

void formant_proc(M_P m)
{
    double p1, p2, p3, p4;
    int    ret, type;
    FO_P f_ptr;

    /* check number of args */
    if ( argc < 2 ) errex(m, 1, "Incorrect number of arguments.");

    /* check type */
    if      ( ! strcmp(argsv[1], "cos"      ) ) type = F_COS;
    else if ( ! strcmp(argsv[1], "triangle" ) ) type = F_TRI;
    else if ( ! strcmp(argsv[1], "highpass" ) ) type = F_HPF;
    else if ( ! strcmp(argsv[1], "lowpass"  ) ) type = F_LPF;
    else if ( ! strcmp(argsv[1], "highshelf") ) type = F_HS;
    else if ( ! strcmp(argsv[1], "lowshelf" ) ) type = F_LS;
    else if ( ! strcmp(argsv[1], "band"     ) ) type = F_BAND;
    else if ( ! strcmp(argsv[1], "graph"    ) ) type = F_GRAPH;
    else errex(m, 1, "Unknown formant type.");

    /* check p1 */
    if ( type != F_GRAPH ) {
	ret = aconvd(argsv[2], &p1);
	if ( ret ) errex(m, 1, "P1 bad number.");
	if ( p1 <  1.0 || p1 > 20000.0 )
	  errex(m, 1, "P1 not between 1.0 and 20,000.0 .");
    }
    else {
	if ( argc > 2 ) errex(m, 1, "Too many arguments for this type.");
	p1 = 0.0; p2 = 0.0; p3 = 0.0; p4 = 0.0;
    }

    /* check p2 */
    if ( type != F_GRAPH ) {
	ret = aconvd(argsv[3], &p2);
	if ( ret ) errex(m, 1, "P2 bad number.");
	switch (type) {
	    case F_COS: case F_TRI: case F_HS: case F_LS:
		if ( p2 < -90.0 || p2 > 90.0 )
		  errex(m, 1, "P2 not between -90.0 and 90.0 .");
		break;
	    case F_HPF: case F_LPF:
		if ( p2 == 0.0 )
		  errex(m, 1, "P2 may not be 0.0 .");
		break;
	    case F_BAND:
		if ( p2 <  1.0 || p2 > 20000.0 )
		  errex(m, 1, "P2 not between 1.0 and 20,000.0 .");
		if ( p1 > p2 )
		  errex(m, 1, "P1 must be <= P2.");
		break;
	}
    }
    else {
	if ( argc > 3 ) errex(m, 1, "Too many arguments for this type.");
	p2 = 0.0; p3 = 0.0; p4 = 0.0;
    }

    /* check p3, if needed */
    if ( type == F_COS || type == F_TRI || type == F_HS || type == F_LS ||
      type == F_BAND ) {
	if ( argc < 5 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(argsv[4], &p3);
	if ( ret ) errex(m, 1, "P3 bad number.");
	switch (type) {
	    case F_COS:  case F_TRI:
		if ( p3 < 0.0 || p3 > 10.0 )
		  errex(m, 1, "P3 not between 0.0 and 10.0 .");
		break;
	    case F_HS:  case F_LS:
		if ( p3 == 0.0 )
		  errex(m, 1, "P3 may not be 0.0 .");
		if ( ( p2 < 0 && p3 > 0 ) || ( p2 > 0 && p3 < 0 ) )
		  errex(m, 1, "p2 and p3 must have same sign.");
		break;
	    case F_BAND:
		if ( p3 < -90.0 || p3 > 90.0 )
		  errex(m, 1, "P3 must be between -90.0 and 90.0 .");
		break;
	}
    }
    else if ( argc > 4 ) errex(m, 1, "Too many arguments for this type.");
    else p3 = 0.0; /* not used */

    /* check p4, if needed */
    if ( type == F_BAND ) {
	if ( argc < 6 ) errex(m, 1, "Incorrect number of arguments.");
	ret = aconvd(argsv[5], &p4);
	if ( ret ) errex(m, 1, "P4 bad number.");
	switch (type) {
	    case F_BAND:
		if ( p4 == 0.0 )
		  errex(m, 1, "P4 may not be 0.0 .");
		break;
	}
    }
    else if ( argc > 5 ) errex(m, 1, "Too many arguments for this type.");
    else p4 = 0.0; /* not used */

    /* input ok */

    /* allocate new formant struct and put in list */
    f_ptr = (FO_P) malloc(sizeof(FO_S));
    (m->d_malloc)++;
    if ( f_ptr == NULL ) errex(m, 1,"Unable to allocate formant struct.");
    if ( m->fo_back == NULL ) m->ws_back->fo_list = f_ptr;
    else m->fo_back->next = f_ptr;
    m->fo_back = f_ptr;

    /* set data in structure */
    f_ptr->type    = type;
    f_ptr->p1      = p1;
    f_ptr->p2      = p2;
    f_ptr->p3      = p3;
    f_ptr->p4      = p4;
    f_ptr->gp_list = NULL;
    f_ptr->next    = NULL;
}

/***********************************************************************

split_proc

***********************************************************************/

void split_proc(M_P m)
{
    int ret, notes;
    WP_P wp_ptr;

    /* check number of args */
    if ( argc != 2 ) errex(m, 1, "Incorrect number of arguments.");

    /* check for optional note list */
    if ( ! strcmp(argsv[1], "list") ) {
	notes = 0;	/* list must follow */
    }
    else { /* no list, need notes per sample to generate */
	ret = aconvi(argsv[1], &notes);
	if ( ret ) errex(m, 1, "Notes bad number.");
	if ( notes < 1 || notes > 128 )
	  errex(m, 1, "Notes must be between 1 and 128.");
    }
    
    /* input ok */

    /* allocate new wave split structure and put in list */
    wp_ptr = (WP_P) malloc(sizeof(WP_S));
    (m->d_malloc)++;
    if ( wp_ptr == NULL ) errex(m, 1,"Unable to allocate split struct.");
    if ( m->wp_back == NULL ) m->ws_back->wp_list = wp_ptr; /* start list */
    else m->wp_back->next = wp_ptr;			   /* continue list */
    m->wp_back = wp_ptr;

    /* initialize data for new wave split structure */
    wp_ptr->notes     = notes;			/* notes / sample  */
    wp_ptr->note_list = NULL;			/* note list	   */
    wp_ptr->wpc_list  = NULL;			/* condition list  */
    wp_ptr->next      = NULL;			/* next split      */
}

/***********************************************************************

note_proc

***********************************************************************/

void note_proc(M_P m)
{
    int ret, center, low, high;
    N_P nl_ptr;

    /* check number of args */
    if ( argc != 4 ) errex(m, 1, "Incorrect number of arguments.");

    /* check for center */
    ret = aconvi(argsv[1], &center);
    if ( ret ) errex(m, 1, "Center bad number.");

    /* check for low */
    ret = aconvi(argsv[2], &low);
    if ( ret ) errex(m, 1, "Low bad number.");

    /* check for high */
    ret = aconvi(argsv[3], &high);
    if ( ret ) errex(m, 1, "High bad number.");

    /* check relations in line */
    if ( ! ( 0 <= low && low <= center && center <= high && high <= 127 ) )
	errex(m, 1, "Must have 0 <= low <= center <= high <= 127.");

    /* check relation to previous, if any, line */
    if ( m->n_back != NULL ) {
	if ( low != m->n_back->high + 1 ) 
	    errex(m, 1, "Must have current low = previous high + 1.");
    }

    /* allocate new note structure and put in list */
    nl_ptr = (N_P) malloc (sizeof(N_S));
    (m->d_malloc)++;
    if ( nl_ptr == NULL ) errex(m, 1, "Unable to allocate note struct.");
    if ( m->n_back == NULL ) m->wp_back->note_list = nl_ptr;
    else m->n_back->next = nl_ptr;
    m->n_back = nl_ptr;

    /* initialize data */
    nl_ptr->center = center;
    nl_ptr->low    = low;
    nl_ptr->high   = high;
    nl_ptr->next   = NULL;
}

/***********************************************************************

condition_proc

***********************************************************************/

void condition_proc(M_P m)
{
    int ret, rel, freq_flag, h, d, r;
    double freq;
    WPC_P wpc_ptr;

    /* check number of args */
    if ( argc < 2 ) errex(m, 1, "Too few arguments.");
    if ( argc > 4 ) errex(m, 1, "Too many arguments.");

    /* check operator */
    if      ( ! strcmp(argsv[1], "="  ) ) rel = SP_C_EQ;
    else if ( ! strcmp(argsv[1], "<"  ) ) rel = SP_C_LT;
    else if ( ! strcmp(argsv[1], "<=" ) ) rel = SP_C_LE;
    else if ( ! strcmp(argsv[1], ">"  ) ) rel = SP_C_GT;
    else if ( ! strcmp(argsv[1], ">=" ) ) rel = SP_C_GE;
    else if ( ! strcmp(argsv[1], "%"  ) ) rel = SP_C_MOD;
    else if ( ! strcmp(argsv[1], "not") ) rel = SP_C_NOT;
    else if ( ! strcmp(argsv[1], "and") ) rel = SP_C_AND;
    else if ( ! strcmp(argsv[1], "or" ) ) rel = SP_C_OR;
    else
	errex(m, 1,
	  "Operator must be one of =, <, <=, >, >=, %, not, and, or .");

    /* comparison operator checks */
    if ( rel == SP_C_EQ ||
	 rel == SP_C_LT ||
	 rel == SP_C_LE ||
	 rel == SP_C_GT ||
	 rel == SP_C_GE ) {

	/* check (optional) frequency flag for comparison */
	if (argc == 4 ) {
	    if ( ! strcmp(argsv[3], "frequency") ) freq_flag = 1;
	    else errex(m, 1, "Expecting frequency.");
	}
	else freq_flag = 0;

	/* check comparison value */
	if ( freq_flag ) {
	    ret = aconvd(argsv[2], &freq);
	    if ( ret ) errex(m, 1, "Frequency bad number.");
	    if ( freq < 1.0 || freq > 20000.0 )
	      errex(m, 1, "Frequency must be between 1.0 and 20000.0 .");
	}
	else {
	    ret = aconvi(argsv[2], &h);
	    if ( ret ) errex(m, 1, "Harmonic bad number.");
	    if ( h < 0 ) errex(m, 1, "Harmonic must be >= 0 .");
	}
    }

    /* modulus operator checks */
    else if ( rel ==SP_C_MOD ) {

	/* check number of args */
	if ( argc != 4 ) errex(m, 1, "Incorrect number of arguments.");

	/* check divisor for modulus */
	ret = aconvi(argsv[2], &d);
	if ( ret ) errex(m, 1, "Divisor bad number.");
	if ( d <= 1 ) errex(m, 1, "Divisor must be greater than 1.");

	/* check remainder for modulus */
	ret = aconvi(argsv[3], &r);
	if ( ret ) errex(m, 1, "Remainder bad number.");
	if ( r < 0 || r >= d )
	  errex(m, 1, "Remainder must be 0 to 1 less than divisor.");
    }

    /* logical operators */
    else {

	/* check number of arguments */
	if ( argc != 2 ) errex(m, 1, "Incorrect number of arguments.");
    }

    /* input ok */

    /* allocate new wave split condition structure and put in list */
    wpc_ptr = (WPC_P) malloc (sizeof(WPC_S));
    (m->d_malloc)++;
    if ( wpc_ptr == NULL )
      errex(m, 1, "Unable to allocate wave split condition struct.");
    if ( m->wpc_back == NULL ) m->wp_back->wpc_list = wpc_ptr; /* start list */
    else m->wpc_back->next = wpc_ptr;			/* continue list */
    m->wpc_back = wpc_ptr;

    /* initialize data for new wave split condition structure */
    wpc_ptr->rel = rel;				/* relation */
    if ( rel == SP_C_EQ ||
	 rel == SP_C_LT ||
	 rel == SP_C_LE ||
	 rel == SP_C_GT ||
	 rel == SP_C_GE ) {
	if ( freq_flag ) {
	    wpc_ptr->harmonic = -1;		/* harmonic  - not used */
	    wpc_ptr->freq = freq;		/* frequency - used     */
	}
	else {
	    wpc_ptr->harmonic = h;		/* harmonic  - used     */
	    wpc_ptr->freq = -1.0;		/* frequency - not used */
	}
	wpc_ptr->divisor = -1;			/* divisor   - not used */
	wpc_ptr->remainder = -1;		/* remainder - not used */
    }
    else if ( rel == SP_C_MOD ) {
	wpc_ptr->harmonic = -1;			/* harmonic  - not used */
	wpc_ptr->freq = -1.0;			/* frequency - not used */
	wpc_ptr->divisor = d;			/* divisor   - used     */
	wpc_ptr->remainder = r;			/* remainder - used     */
    }
    else {
	wpc_ptr->harmonic  = -1;		/* not used */
	wpc_ptr->freq      = -1;		/* not used */
	wpc_ptr->divisor   = -1;		/* not used */
	wpc_ptr->remainder = -1;		/* not used */
    }
    wpc_ptr->next = NULL;			/* next condition       */
}

/***********************************************************************

wave_proc

***********************************************************************/

void wave_proc(M_P m)
{
    char *wavnos[] = { "1", "2", "3", "4" };
    W_P ptr;

    /* check number of args */
    if ( argc != 1 ) errex(m, 1, "Incorrect number of arguments.");

    /* check number of waves for voice and file */
    if ( m->v_waves+1 > S1M_WPV ) errex(m, 1, "Too many waves in voice.");
    if ( m->f_waves+1 > S1M_WPF ) errex(m, 1, "Too many waves in file.");

    /* input ok */

    /* increment wave counts for voice and file */
    m->v_waves++;
    m->f_waves++;

    /* allocate new wave structure and put in list */
    ptr = (W_P) malloc (sizeof(W_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 1, "Unable to allocate wave struct.");
    if ( m->w_back == NULL ) m->w_list = ptr;
    else m->w_back->next = ptr;
    m->w_back = ptr;
    
    /* set data in structure */
    ptr->no      = m->f_waves;			/* wave number   */
    strcpy(ptr->cat, m->v_cat);			/* wave category */
    strncpy(ptr->name, m->v_name, 5);		/* wave name     */
    ptr->name[5] = '\0';
    strcat(ptr->name, wavnos[m->v_waves-1]);
    ptr->next    = NULL;			/* next wave     */
    ptr->s_list  = NULL;			/* sample list   */
}

/***********************************************************************

sample_proc

***********************************************************************/

void sample_proc(M_P m)
{
    int ret, center, low, high, snlen;
    double freq;
    char hconv[] = "0123456789abcdef";
    S_P ptr;
    double per;

    /* check number of args */
    if ( argc < 4 ) errex(m, 1, "Too few arguments.");
    if ( argc > 5 ) errex(m, 1, "Too many arguments.");

    /* check number of samples per wave and file */
    if ( m->w_samples+1>S1M_SPW ) errex(m,1,"Too many samples in wave.");
    if ( m->f_samples+1>S1M_SPF ) errex(m,1,"Too many samples in file.");

    /* check center */
    ret = aconvi(argsv[1], &center);
    if ( ret ) errex(m, 1, "Center bad number.");
    if ( center<0 || center>127 ) errex(m,1,"Center must be >= 0 and <= 127.");
    
    /* check low */
    ret = aconvi(argsv[2], &low);
    if ( ret ) errex(m, 1, "Low bad number.");
    if ( low<0 || low>center ) errex(m, 1, "Low must be >= 0 and <= center.");

    /* check high */
    ret = aconvi(argsv[3], &high);
    if ( ret ) errex(m, 1, "High bad number.");
    if ( high < center || high > 127 )
        errex(m, 1, "High must be >= center and <= 127.");

    /* check frequency, or calculate */
    if ( argc == 5 ) {		/* frequency specified */
	ret = aconvd(argsv[4], &freq);
	if ( ret ) errex(m, 1, "Frequency bad number.");
	if ( freq < m->o_lowfreq || freq > S1M_SR / 2.0 )
	    errex(m, 1, "Frequency out of range\n");
    }
    else {				/* calculate frequency */
	freq = 440.0 * pow(pow(2.0, 1.0 / 12.0), center - 69);
    }

    /* input ok */

    /* allocate new sample structure and put in list */
    ptr = (S_P) malloc (sizeof(S_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 1, "Unable to allocate sample struct.");
    if ( m->s_back == NULL ) m->w_back->s_list = ptr;
    else (m->s_back)->next = ptr;
    m->s_back = ptr;
    
    /* update m */
    m->w_samples += 1;
    m->f_samples += 1;
    m->h_no = 0;

    /* set data in structure */
    ptr->no = m->f_samples;		/* sample number */

    strcpy(ptr->cat, m->w_back->cat);   /* copy sample category from wave */

    strcpy(ptr->name, m->w_back->name);	/* generate sample name */
    snlen = strlen(ptr->name);
    ptr->name[snlen    ] = hconv[(center >> 4) & 0x0f]; /* append hex of */
    ptr->name[snlen + 1] = hconv[ center       & 0x0f]; /* center */
    ptr->name[snlen + 2] = '\0';

    ptr->freq	= freq;
    ptr->center	= center;
    ptr->low	= low;
    ptr->high	= high;

    /* additional calculations for sample generation */
    per = S1M_SR / freq;
    ptr->loop_len = (int) floor(per) + 1;
    ptr->loop_fract = 64 - (int) round((per - floor(per)) * 64);
    if ( ptr->loop_fract == 64 ) {
	ptr->loop_fract = 0;
	ptr->loop_len -= 1;
    }
    ptr->smpl_len = ptr->loop_len + 2;

    /* initialize pointers to next (if any) structure */
    ptr->next   = NULL;
    ptr->h_list = NULL;
}

/***********************************************************************

harmonic_proc

***********************************************************************/

void harmonic_proc(M_P m)
{
    int ret, hno;
    double level, phase, freq;
    H_P ptr;

    /* check number of args */
    if ( argc < 3  || argc > 4 ) errex(m, 1, "Incorrect number of arguments.");

    /* check harmonic number */
    ret = aconvi(argsv[1], &hno);
    if ( ret ) errex(m, 1, "Harmonic bad number.");
    if ( hno <= m->h_no ) errex(m, 1, "Number must > 0 and increasing.");
    freq = m->s_back->freq * hno *
      pow(pow(2.0, 1.0/12.0),(m->s_back->high - m->s_back->center)+m->v_bend);
    if ( freq > S1M_SR / 2.0 ) {
	fprintf(m->log, "Warning (line %d):  Harmonic too high, may alias.\n",
	  m->linecnt);
    }

    /* check level */
    ret = aconvd(argsv[2], &level);
    if ( ret ) errex(m, 1, "Level bad number.");
    if ( level < -90.0 || level > 0.0 )
	errex(m, 1, "Level must be between -90.0 and 0.0 (dba).");

    /* check phase */
    if ( argc == 4 ) {	/* phase specified */
	ret = aconvd(argsv[3], &phase);
	if ( ret ) errex(m, 1, "Phase bad number.");
	if ( phase < 0.0 || phase >= 1.0 )
	    errex(m, 1, "Phase must be >= 0.0 and < 1.0 .");
    }
    else phase = 0.0; /* phase not specified, set to 0.0 */

    /* input ok */

    /* allocate new harmonic structure and put in list */
    ptr = (H_P) malloc (sizeof(H_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 1, "Unable to allocate harmonic struct.");
    if ( m->h_back == NULL ) m->s_back->h_list = ptr;
    else (m->h_back)->next = ptr;
    m->h_back = ptr;
    
    /* set data in structure */
    m->h_no    = hno;
    ptr->no    = m->h_no;
    ptr->level = level;
    ptr->phase = phase;
    ptr->next  = NULL;
}

/* end of input.c */

