/**************************************************************

s1m.c - functions directly related to s1m files

This file is a part of sscsss -- Simple Single Cycle Sample Set Synthesizer.

Copyright (C) 2002 David Baker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


intended usage

	open s1m file
	call init_s1m_header
	call write_s1m_header
	some combination of
		call add_s1m_wave
		call add_s1m_layer
		call add_s1m_sample
		call write_s1m_sample
	call write_s1m_end
	reset file pointer to beginning
	call write_s1m_header (updated header)
	close s1m file

**************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "sscsss.h"
#include "s1m.h"
#include "proto.h"

/* private global variables */

static int w_offset = 0;		/* offset for next wave or layer */
static unsigned char hdr[SDATOFFSET];	/* buffer for s1m header */

/**************************************************************

init_header - create empty/initialized s1m header

**************************************************************/

/* read an empty s1m file in for initialization */
void init_s1m_header(M_P m, const char path[])
{
    FILE *f;
    size_t incnt;

    /* open template file init.s1m */
    f = fopen(path, "rb");
    if ( f == NULL ) errex(m, 0, "Cannot open init.s1m for reading.");

    /* read template file into hdr */
    incnt = fread(hdr, SDATOFFSET, 1, f);
    if ( ferror(f) ) errex(m, 0, "Failed to read init.s1m.");

    fclose(f);
}

/**************************************************************

add_wave - add a wave parameter record to s1m header

**************************************************************/

void add_s1m_wave(M_P m, W_P wp)
{
    int offset = WAVEPARAMS + w_offset;
    int i;

    /* store offset to this wave record */
    cp_rev_bytes(m, (void *)&(w_offset),
      (void *)&hdr[WAVEOFFSET+(4*(wp->no-1))], 4);

    /* store name */
    for ( i = 0; i < (int) strlen(wp->name) ; i++ ) {
	hdr[offset+WNAME+i] = wp->name[i];
    }
    for ( i = strlen(wp->name); i < 8 ; i++ ) {
	hdr[offset+WNAME+i] = ' ';
    }
    hdr[offset+WNAME+8] = 0x00;

    /* store number of layers (= samples for this program) */
    hdr[offset+WLAYERS] = (unsigned char) wp->layers;
    hdr[offset+WLAYERS+1] = 0x00;

    /* store category */
    hdr[offset+WCATEGORY] = (unsigned char) chkcat(wp->cat);

    /* update offset */
    w_offset += 12;
}

/**************************************************************

add_layer - add a wave/layer parameter record to s1m header

**************************************************************/

void add_s1m_layer(M_P m, S_P sp)
{
    int offset = WAVEPARAMS + w_offset;
    short t;

    /* store sample number and bank (RAM) */
    t = (short) sp->no - 1;
    cp_rev_bytes(m, (void *) &t, (void *) &hdr[offset+LSMPLNO], 2);
    hdr[offset+LSMPLNO] |= 0x40;

    /* store level, 0 is max ?! */
    hdr[offset+LLEVEL] = 0;

    /* store ?? */
    hdr[offset+LMYS1] = 0;

    /* store low note */
    hdr[offset+LLOWNOTE] = (unsigned char) sp->low;

    /* store high note */
    hdr[offset+LHIGHNOTE] = (unsigned char) sp->high;

    /* store low velocity */
    hdr[offset+LLOWVEL] = 1;

    /* store high velocity */
    hdr[offset+LHIGHVEL] = 127;

    /* store pan and frequency mode */
    hdr[offset+LPAN] = 0x40 | 0x00;  /* pan center and ratio freq mode */

    /* store coarse tune */
    hdr[offset+LCOARSETUNE] = 0x40;

    /* store fine tune */
    hdr[offset+LFINETUNE] = 0x40;

    /* store sample play */
    hdr[offset+LSMPLPLAY] = 0x20; /* default from sample */

    /* store ?? */
    hdr[offset+LMYS2  ] = 0;
    hdr[offset+LMYS2+1] = 0;

    /* store sample start offset */
    hdr[offset+LSTARTOFF  ] = 0;
    hdr[offset+LSTARTOFF+1] = 0;

    /* store sample length offset */
    hdr[offset+LLENGTHOFF  ] = 0;
    hdr[offset+LLENGTHOFF+1] = 0;

    /* update offset */
    w_offset += 18;
}

/**************************************************************

add_sample - add a sample parameter record to s1m header

**************************************************************/

void add_s1m_sample(M_P m, S_P sp)
{
    int bufoffset = SHDROFFSET + ( 30 * (sp->no - 1));
    int sizoffset = SSIZOFFSET + (  4 * (sp->no - 1));
    int i, enddatamendloop, t1;
    short t2;

    /* store name */
    for ( i = 0; i < (int) strlen(sp->name) ; i++ ) {
	hdr[bufoffset + SNAME + i] = sp->name[i];
    }
    for ( i = strlen(sp->name); i < 8 ; i++ ) {
	hdr[bufoffset + SNAME + i] = ' ';
    }

    /* store category */
    hdr[bufoffset + SCAT] = (unsigned char) chkcat(sp->cat);

    /* store some constants */
    hdr[bufoffset + S09] = '\0';
    hdr[bufoffset + S0A] = (unsigned char) '\374'; /* 0xfc */
    hdr[bufoffset + SMYS1    ] = 0;
    hdr[bufoffset + SMYS1 + 1] = 0;

    /* store fraction */
    hdr[bufoffset + SFRACT] = (unsigned char) sp->loop_fract;

    /* store top - start */
    hdr[bufoffset + STOPMSTART    ] = 0;
    hdr[bufoffset + STOPMSTART + 1] = 0;
    hdr[bufoffset + STOPMSTART + 2] = 0;

    /* store tune */
    hdr[bufoffset + STUNE] = 0;

    /* store loop length */
    cp_rev_bytes(m, (void *)&sp->loop_len, (void *)&hdr[bufoffset+SLEN], 3);

    /* store center */
    hdr[bufoffset + SCENTER] = (unsigned char) sp->center;

    /* store start */
    hdr[bufoffset + SSTART    ] = 0;
    hdr[bufoffset + SSTART + 1] = 0;
    hdr[bufoffset + SSTART + 2] = 0;

    /* store end of data - end of loop */
    enddatamendloop = sp->smpl_len - sp->loop_len;
    cp_rev_bytes(m, (void *) &enddatamendloop,
      (void *) &hdr[bufoffset + SENDDATAMENDLOOP], 3);

    /* store sample order */
    t2 = sp->no - 1;
    cp_rev_bytes(m, (void *) &t2, (void *) &hdr[bufoffset + SORDER], 2);

    /* store sample size (bytes) in size array */
    t1 = sp->smpl_len * 2;
    cp_rev_bytes(m, (void *) &t1, (void *) &hdr[sizoffset], 4);

}

/**************************************************************

write_header - write s1m header to file

**************************************************************/

void write_s1m_header(M_P m)
{
    size_t out;
    short t1;

    /* store wave parameter offset */
    t1 = (short) w_offset;
    cp_rev_bytes(m, (void *)&(t1), (void *)&hdr[NEXTWAVE], 2);

    /* write out header */
    out = fwrite(hdr, SDATOFFSET, 1, m->outfile);
    if ( out != 1 ) errex(m, 0, "Failed to write s1m header.");
}

/**************************************************************

write_sample - write s1m sample data to file

**************************************************************/

void write_s1m_sample(M_P m, S_P s, double sample[], short isample[])
{
    int i;
    short ipoint;
    size_t out;

    /* convert samples from double to short */
    for ( i = 0; i < s->smpl_len; i++ ) {
	ipoint = (short) round(32767 * sample[i]);
	cp_rev_bytes(m, (void *)&ipoint, (void *)&isample[i], 2);
    }
    
    /* write sample data to end of s1m file */
    out = fwrite(isample, s->smpl_len * 2, 1, m->outfile);
    if ( out != 1 ) errex(m, 0, "Failed to write sample data.");

    /* update sample bytes for file */
    m->f_size += s->smpl_len * 2;
}

/**************************************************************

write_end - finish off s1m file

**************************************************************/

void write_s1m_end(M_P m)
{
    size_t out;
    unsigned char zeroes[128];
    long data_end;
    int i;

    data_end = ftell(m->outfile);
    if ( data_end == -1 ) errex(m, 0, "Failed to get file position.");

    if ( data_end % 128 != 0 ) {
	for ( i = 0; i < 128; i++ ) zeroes[i] = '\0';
	out = fwrite(zeroes, 128 - (data_end % 128), 1, m->outfile);
	if ( out != 1 ) errex(m, 0, "Failed to write end of s1m file.");
    }
}

/***********************************************************************

chkcat

***********************************************************************/

int chkcat(char cat[])
{
    int found, i;
    char *cats[] = { "--", "Pf", "Cp", "Or", "Gt", "Ba", "St", "En", "Br",
      "Rd", "Pi", "Ld", "Pd", "Fx", "Et", "Pc", "Se", "Dr", "Sc", "Vo",
      "Co", "Wv", "Sq" };
    found = -1;
    for ( i = 0; i < 23; i++ ) {
	if ( ! strcmp(cat, cats[i]) ) {
	    found = i;
	    break;
	}
    }
    return found;
}

/* end of s1m.c */

